# -*- coding: utf-8 -*-
"""
    __init__

    Python API for prestashop

    :copyright: © 2013 by Openlabs Technologies & Consulting (P) Limited
    :license: BSD, see LICENSE for more details.
"""
from .api import PrestaShopWebservice, PrestaShopWebserviceException    # noqa



# -*- coding: utf-8 -*-
"""
    api


    :copyright: © 2013 by Openlabs Technologies & Consulting (P) Limited
    :license: BSD, see LICENSE for more details.
"""
import functools

import requests
from lxml import objectify, etree


class PrestaShopWebserviceException(Exception):
    """
    Abstract Exception class for all exceptions
    """
    pass


class PrestaShopWebservice(object):
    """
    Web service client for prestashop

    Use of this API requires an authentication key. If you do not have one,
    you can generate one from your Back Office and click on the
    "Tools / Web Service" tab.

    `Read More <http://doc.prestashop.com/display/PS14/\
            Chapter+1+-+Creating+Access+to+Back+Office>`_
    """

    def __init__(self, url, key, debug=False):
        """
        :param url: The store's root path (ex: http://store.com/ )
        :param key: The authentication key (ex: ZR92FNY5UFRERNI3O9Z5QDHWKTP3Y)
        :param debug: A Boolean indicating whether the Web service must use
                      its debug mode
        """
        self.url = url
        if self.url.endswith('/'):
            self.url = self.url[:-1]
        self.key = key
        self.debug = debug

    @property
    def session(self):
        if not hasattr(self, '_session'):
            self._session = requests.Session()
            self._session.auth = (self.key, 'ignore')
        return self._session

    def __getattr__(self, name):
        """
        Return a Resource proxy object for the attribute
        """
        return type(
            'PrestaShopWebservice.' + name,
            (ResourceProxy,),
            {
                '__resource__': name,
                'session': self.session,
                'url': '{0}/api/{1}'.format(self.url, name)
            }
        )


def singular(function):
    @functools.wraps(function)
    def wrapper(cls, *args, **kwargs):
        rv = function(cls, *args, **kwargs)
        return rv.getchildren()[0]
    return wrapper


class ResourceProxy(object):
    """
    Class to represent any resource on prestashop
    """

    @classmethod
    def check_status(cls, response):
        """
        Check the response status code and if it is not one of the
        HTTP OK codes, raise an exception with the content of the response

        :param response: response object from requests
        """
        if response.status_code not in (200, 201):
            raise PrestaShopWebserviceException(response.content)

    @classmethod
    def wrap_in_prestashop_tag(cls, child):
        """
        Wrap the given Element child into a prestashop root
        tag and deannotate elements.

        :param child: lxml Element
        """
        root = objectify.Element('prestashop')
        root.insert(0, child)
        objectify.deannotate(root)
        return root

    @classmethod
    @singular
    def create(cls, xml):
        """
        Creates a record on the server
        """
        root = cls.wrap_in_prestashop_tag(xml)
        response = cls.session.post(
            cls.url, data={'xml': etree.tostring(root)}
        )
        cls.check_status(response)

        return objectify.fromstring(response.content)

    @classmethod
    @singular
    def get_schema(cls):
        """
        Returns a blank xml
        """
        response = cls.session.get(cls.url + '?schema=blank')
        cls.check_status(response)

        return objectify.fromstring(response.content)

    @staticmethod
    def make_params(display=None, filters=None,
                    sort=None, limit=None, offset=None, date=None):
        """
        Builds the URL paramters as a dictionary

        :param display: A list of field names
                        eg: ['firstname', 'lastname']
        :param filters: A dictionary of field name and values
                        eg: {'firstname': 'sharoon'}
        :param sort: a list of tuple of field and value
                     eg: [('firstname', 'ASC'), ('lastname', 'DESC')]
        :param limit: The number of records to be returned
        :param offset: The number of records to skip
        :param date: This should be a logically True value to have a date based
                     filter. For details,
                     `refer <http://www.prestashop.com/forums/topic/\
                             101502-webservice-api-filter-for-date-ranges/>`_
        """
        params = {}

        if display:
            params['display'] = display if display == 'full' \
                else '[' + ','.join(display) + ']'

        if filters:
            for key, value in filters.iteritems():
                params['filter[%s]' % key] = '[%s]' % value

        if sort:
            params['sort'] = '[' + ','.join([
                '{0}_{1}'.format(field, order) for field, order in sort
            ]) + ']'

        if limit and not offset:
            params['limit'] = limit
        elif limit and offset:
            params['limit'] = '%d,%d' % (offset - 1, limit)

        if date:
            params['date'] = 1

        return params

    @classmethod
    def get_list(cls, as_ids=False, display=None,
                 filters=None, sort=None, limit=None, offset=None, date=None):
        """
        Gets a list of records by sending GET on the Collection
        URI of the resource.

        :param as_ids: If True a list of ids are returned, or the
                       XML object is returned
        :param display: A list of fields that needs to be returned.
                        By default only the ID is fetched
        :param filters: A dictionary of field name and values
                        eg: {'firstname': 'sharoon'}
        :param sort: a list of tuple of field and value
                     eg: [('firstname', 'ASC'), ('lastname', 'DESC')]
        :param limit: The number of records to be returned
        :param offset: The number of records to skip
        :param date: This should be a logically True value to have a date based
                     filter.
        """
        if display is None:
            display = []

        if as_ids and 'id' not in display and display != 'full':
            display.append('id')

        params = cls.make_params(display, filters, sort, limit, offset, date)
        response = cls.session.get(cls.url, params=params)
        cls.check_status(response)
        rv = objectify.fromstring(response.content)

        if as_ids:
            return map(
                lambda r: int(r.id),
                getattr(rv, cls.__resource__).iterchildren()
            )

        return getattr(rv, cls.__resource__).getchildren()

    @classmethod
    @singular
    def get(cls, id):
        """
        Reads a single record from the server

        :param id: Id of the record to read
        """
        response = cls.session.get('%s/%d' % (cls.url, id))
        cls.check_status(response)
        return objectify.fromstring(response.content)

    @classmethod
    def update(cls, id, xml):
        """
        Updates a single record on the server
        """
        root = cls.wrap_in_prestashop_tag(xml)

        response = cls.session.put(
            '%s/%d' % (cls.url, id),
            data=etree.tostring(root)
        )
        cls.check_status(response)

        return objectify.fromstring(response.content)

    @classmethod
    def delete(cls, id):
        """
        Deletes a single record from the server

        :return: True if deletion was successful
        """
        response = cls.session.delete('%s/%d' % (cls.url, id))
        cls.check_status(response)
        return response.status_code == requests.codes.ok



# -*- coding: utf-8 -*-
"""
    api


    :copyright: © 2013 by Openlabs Technologies & Consulting (P) Limited
    :license: BSD, see LICENSE for more details.
"""
import os
from collections import namedtuple
import pkg_resources

from pystashop.api import PrestaShopWebservice, singular, ResourceProxy

Response = namedtuple('Response', 'status_code, content')


class FakeSession(object):
    """
    A requests.Session like class which just checks the filesystem folder
    for a possible xml to return
    """
    def __init__(self, version):
        self.version = version

    @property
    def folder(self):
        """
        Return the folder where the xml files are stored
        """
        root_xml_folder = pkg_resources.resource_filename('mockstashop', 'xml')

        # XXX: May be handle minor versions if there are API changes inside
        # minor releases ?
        return os.path.join(root_xml_folder, self.version)

    @staticmethod
    def get_path(url):
        """
        Return just the path of the url

        >>> FakeSession.get_path('/api/customers')
        'customers.xml'
        >>> FakeSession.get_path('/api/customers/1')
        'customers/1.xml'
        >>> FakeSession.get_path('/api/orders')
        'orders.xml'

        :param url: URL of the file
        """
        # Different resources like customers, orders as each of them have files
        # names like 1.xml, 2.xml for individual records.
        # If the last part of URL is an integer, return last two parts of URL
        first, second_last, last = url.rsplit('/', 2)

        if last.isdigit():
            return os.path.join(second_last, last + '.xml')
        else:
            # If the last part is not an integer, return last part itself
            return last + '.xml'

    def build_response_for(self, path):
        """
        Build a mock http response with code and content

        :param path: Path to file inside self.folder
        """
        filename = os.path.join(self.folder, path)
        if os.path.exists(filename):
            return Response(
                status_code=200,
                content=open(filename).read()
            )
        else:
            return Response(
                status_code=404,
                content=open(os.path.join(self.folder, 'error.xml')).read()
            )

    def get(self, url, params=None):
        """
        A GET request is made to the URL

        :param url: URL of the file
        :param params: Parameters used for filtering of records.
                       This is not implemented yet.
        """
        # TODO: handle params
        path = self.get_path(url)
        return self.build_response_for(path)


class MockstaShopWebservice(PrestaShopWebservice):
    """
    A mock service client for prestashop


    `Read More <http://doc.prestashop.com/display/PS14/\
                Chapter+1+-+Creating+Access+to+Back+Office>`_
    """

    def __init__(self, url, key, debug=False):
        """
        :param url: The store's root path (ex: http://store.com/ )
        :param key: The authentication key (ex: ZR92FNY5UFRERNI3O9Z5QDHWKTP3Y)
        :param debug: A Boolean indicating whether the Web service must use
                      its debug mode
        """
        self.url = url
        if self.url.endswith('/'):
            self.url = self.url[:-1]
        self.key = key
        self.debug = debug
        self.version = '1.5'

    @property
    def session(self):
        """Generate a fake requests session
        """
        if not hasattr(self, '_session'):
            self._session = FakeSession(self.version)
            self._session.auth = (self.key, 'ignore')
        return self._session

    def __getattr__(self, name):
        """
        Return a Resource proxy object for the attribute

        :param name: Name of the resource
        """
        return type(
            'MockstaShopWebservice.' + name,
            (MockResourceProxy,),
            {
                '__resource__': name,
                'session': self.session,
                '__version__': self.version,
                'url': '{0}/api/{1}'.format(self.url, name)
            }
        )


class MockResourceProxy(ResourceProxy):
    """
    Class to represent any resource on prestashop
    """

    @classmethod
    @singular
    def create(cls, xml):
        """
        Creates a record on the server

        :param xml: XML for which record needs to be created
        """
        raise Exception('Not Implemented Yet')

    @classmethod
    def update(cls, id, xml):
        """
        Updates a single record on the server

        :param id: ID of the record to be updated
        :param xml: Details to be updated in XML
        """
        raise Exception('Not Implemented Yet')

    @classmethod
    def delete(cls, id):
        """
        Deletes a single record from the server

        :param id: ID of record to be deleted

        :return: True if deletion was successful
        """
        raise Exception('Not Implemented Yet')



# -*- coding: utf-8 -*-
"""
    setup


    :copyright: © 2013 by Openlabs Technologies & Consulting (P) Limited
    :license: BSD, see LICENSE for more details.
"""
import os
from setuptools import setup

execfile(os.path.join('pystashop', 'version.py'))

setup(
    name='pystashop',
    version=VERSION,
    url='https://github.com/openlabs/pystashop/',
    author='Sharoon Thomas, Openlabs Technologies',
    author_email='info@openlabs.co.in',
    description='Prestashop Webservice Python API Client',
    long_description=open('README.rst').read(),
    packages=['pystashop', 'mockstashop'],
    package_dir={'mockstashop': 'mockstashop'},
    package_data={
        'mockstashop': [
            'xml/1.5/*.xml',
            'xml/1.5/shops/*.xml',
            'xml/1.5/customers/*.xml',
            'xml/1.5/addresses/*.xml',
            'xml/1.5/countries/*.xml',
            'xml/1.5/states/*.xml',
            'xml/1.5/currencies/*.xml',
            'xml/1.5/products/*.xml',
            'xml/1.5/combinations/*.xml',
            'xml/1.5/product_option_values/*.xml',
            'xml/1.5/orders/*.xml',
            'xml/1.5/order_details/*.xml',
            'xml/1.5/order_states/*.xml',
            'xml/1.5/languages/*.xml',
        ],
    },
    zip_safe=False,
    platforms='any',
    install_requires=[
        'requests',
        'lxml',
    ],
    classifiers=[
        'Development Status :: 6 - Mature',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    test_suite='tests.suite',
)
